//
//  MMInvocation.m
//
//  Created by Herbert Ptzl on Sun Mar 24 2002.
//  Copyright (c) 2002 H.Ptzl. 
//
//  This file is part of MathMap.
//
//  MathMap is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  MathMap is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with MathMap; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

#import "MMInvocation.h"
#import "MMExpression.h"
#import "MMUserValue.h"

#import "tags.h"
#import "noise.h"

#import <mach/thread_act.h>


@interface MMExpression	(Invocation)

- (mathmap_t *)mathmap;

@end

@implementation MMExpression	(Invocation)

- (mathmap_t *)mathmap		{ return _mathmap; }

@end



enum {
    MMInvocationPrepared = 1,		
    MMInvocationLaunching = 2,		
    MMInvocationActive = 3,		
    MMInvocationDone = 4,
    MMInvocationAborted = 5
};

#define	THREAD_FINISHED(d)	(((d)->state == MMInvocationDone) ||		\
                                 ((d)->state == MMInvocationAborted))

#define	THREAD_RUNNING(d)	(((d)->state == MMInvocationActive) ||		\
                                 ((d)->state == MMInvocationLaunching))


struct _thread_data {
    NSSize bitmapSize;
    unsigned char *bitmapData;
    mathmap_invocation_t *invocation;
    NSRange range;
    unsigned int state;
    BOOL isThreaded;
};



void *	RenderThread(void *data) {
    MMThreadData *threadData = (MMThreadData *)data;
    mathmap_invocation_t *invocation;
    unsigned char *output;
    int cancelType;
    int width, height;
    int first, last, rowstride;
    
    width = threadData->bitmapSize.width;
    height = threadData->bitmapSize.height;
    output = threadData->bitmapData;
    invocation = threadData->invocation;
    first = threadData->range.location;
    last = first + threadData->range.length;
    rowstride = width*4;
    
    threadData->state = MMInvocationActive;    
    if (threadData->isThreaded) {
         pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, &cancelType);
    }    

    call_invocation(invocation, first, last, 
        output + (first*rowstride));

    threadData->state = MMInvocationDone;
    if (threadData->isThreaded) {
         pthread_setcanceltype(cancelType, &cancelType);
         sleep(1000);
    }    
    return threadData;
}



@implementation MMInvocation


- (void)copyUserValues {
    mathmap_invocation_t *invocation = 
        _threadData->invocation;
    mathmap_t *mathmap = [_expression mathmap];
    NSEnumerator *enumerator = 
        [_userValues objectEnumerator];
    MMUserValue *value;
    
    while ((value = [enumerator nextObject]) != nil) {
	userval_info_t *info = 
            lookup_matching_userval(mathmap->userval_infos, 
                [value cInfo]);
	if (info != 0)
	    [value copyToCValue: &invocation->uservals[info->index]];
    }
}


- (id)initWithExpression:(MMExpression *)expression
    userValues:(NSArray *)values
    bitmapImageRep:(NSBitmapImageRep *)imageRep {
    self = [super init];
    if (self) {
        NSSize size = [imageRep size];
        mathmap_t *mathmap = [expression mathmap];
        
        _timeValue = 0.0;
        _flags.antialiasing = NO;
        _flags.supersampling = NO;
        _expression = [expression retain];
        _imageRep = [imageRep retain];
        _userValues = [values copy];
        _threadData = NSZoneMalloc([self zone], sizeof(MMThreadData));
        _threadData->invocation = 
            invoke_mathmap(mathmap, 0, 
            size.width, size.height);
        _threadData->isThreaded = NO;
        _threadData->state = MMInvocationPrepared;
    }
//    NSLog(@"+ Invocation %p", self);
    return self;
}

- (void) dealloc {
    if (THREAD_RUNNING(_threadData)) [self stop];
    [_expression release];
    [_imageRep release];
    [_userValues release];
    free_invocation(_threadData->invocation);
    NSZoneFree([self zone], _threadData);
//    NSLog(@"- Invocation %p", self);
    [super dealloc];
}

- (MMExpression *)expression		{ return _expression; }
- (NSArray *)userValues			{ return _userValues; }
- (NSBitmapImageRep *)imageRep		{ return _imageRep; }

- (void)setTimeValue:(float)value 	{ _timeValue = value; }

- (void)setSupersampling:(BOOL)flag	{ _flags.supersampling = flag; }
- (void)setAntialiasing:(BOOL)flag	{ _flags.antialiasing = flag; }

- (void)renderThreaded:(BOOL)flag 
    range:(NSRange)range {
    mathmap_invocation_t *invocation = _threadData->invocation;

    if (_threadData->state == MMInvocationPrepared)
        [self copyUserValues];
    
    invocation->antialiasing = _flags.antialiasing;
    invocation->supersampling = _flags.supersampling;
    invocation->output_bpp = 4;
    invocation->num_rows_finished = range.location;
    invocation->current_t = _timeValue;
    invocation->origin_x = 0;
    invocation->origin_y = 0;
    invocation->edge_color[0] = 0;
    invocation->edge_color[1] = 0;
    invocation->edge_color[2] = 0;
    invocation->edge_color[3] = 0;

//    init_invocation(invocation);
    update_image_internals(invocation);

    _threadData->bitmapData = [_imageRep bitmapData];
    _threadData->bitmapSize = [_imageRep size];
    _threadData->state = MMInvocationLaunching;
    _threadData->isThreaded = flag;
    _threadData->range = range;

    if (flag) {
        pthread_create(&_thread, NULL, RenderThread, _threadData);
        pthread_detach(_thread);
    } else {
        RenderThread(_threadData);
    }
}

- (void)renderRange:(NSRange)range {
    [self renderThreaded:NO range:range];
}

- (void)render {
    NSSize size = [_imageRep size];
    [self renderRange:NSMakeRange(0, size.height)];
}



- (void)stop {
    if (pthread_cancel(_thread)) {
    //    thread_terminate(pthread_mach_thread_np(_thread));
    }
    thread_terminate(pthread_mach_thread_np(_thread));	/*FIXME*/
    _threadData->state = MMInvocationAborted;
}

- (BOOL)isRunning		{ return THREAD_RUNNING(_threadData); }
- (BOOL)hasFinished		{ return THREAD_FINISHED(_threadData); }

- (int)currentRow { 
    mathmap_invocation_t *invocation = _threadData->invocation;
//    printf("currentRow = %d\n", invocation->num_rows_finished);
    return invocation->num_rows_finished;
}

- (float)progress { 
    NSSize size = [_imageRep size];
    float first = _threadData->range.location;
    if (size.height) 
        return ([self currentRow] - first)/(size.height);
    else return -1.0;
}

- (NSRect)renderedRect { 
    NSSize size = [_imageRep size];
    float first = _threadData->range.location;
    return NSMakeRect(0, first, 
        size.width, [self currentRow] - first);
}

@end
